#define INDENT "   "

#include <iostream>

// LibXML2 Includes
extern "C" {
#include <libxml/parser.h>
#include <libxml/tree.h>
}

using namespace std;

// Get element name or throw exception
string safe_elementName(xmlNodePtr node){
    if(not node->name){
        throw string("XML-Element without name");
    }
    return string((const char*) node->name);
}

// Get attribute or throw exception
string safe_getAttribute(xmlNodePtr node,
                         string attr){
    // Check node exisis
    if(not node) { throw string("Request for attribute in void node"); }
    // Read attribute
    xmlChar *val = xmlGetProp(node, (const xmlChar*) attr.c_str());
    // Check if attribute exists
    if(not val){ throw string("Return element without ") + attr + " attribute"; }
    // Construct return string
    string ret((const char*) val);
    // Free attribute buffer
    xmlFree(val);
    // return attribute value
    return ret;
}

string indent(int tabLevel){
    string ret;
    for(int i = 0; i < tabLevel; i++) { ret += INDENT; }
    return ret;
}    

// Abstract element visitor class
class ElementVisitor {
public:
    // Trigger for enabling a visitor
    virtual bool applicable(xmlNodePtr node) const = 0;
    // Abstract visit method
    virtual void visit(xmlNodePtr node) const = 0;
    // Abstract visitBetween method
    virtual void visitBetween(xmlNodePtr node) const{}
};

// Generic accept method
void accept(xmlNodePtr node, ElementVisitor &visitor){
    // Loop over node/element children
    for(xmlNodePtr child = node->children;
        child;
        child = child->next){
        // If visitor is applicable, visit child element
        if(visitor.applicable(child)) { visitor.visit(child); }
        // If next is applicable, visit in between
        if(child->next and visitor.applicable(child->next)) {
            visitor.visitBetween(child);
        }
    }
}

// Class to visit a method body
class Body : public ElementVisitor {
public:
    // Constructor for 'body' object
    Body(){}
    // Applicable if element name is 'body'
    virtual bool applicable(xmlNodePtr node) const{
        return safe_elementName(node) == "body";
    }
    // Method to visit a 'body' element
    void visit(xmlNodePtr node) const{
        // Skip to text element
        node = node->children;
        // Output body text
        cout << (const char*) node->content;
    }
};

// Class to visit a method formal parameter
class Parameter : public ElementVisitor {
public:
    // Constructor for 'parameter' object
    Parameter(){}
    // Applicable if element name is 'parameter'
    virtual bool applicable(xmlNodePtr node) const{
        return safe_elementName(node) == "parameter";
    }
    // Method to visit a 'parameter' element
    void visit(xmlNodePtr node) const{
        // Output parameter declaration
        cout << safe_getAttribute(node, "type")
             << " "
             << safe_getAttribute(node, "name");
    }
    virtual void visitBetween(xmlNodePtr node) const{
        cout << ", ";
    }
};

// Class to visit a method return type
class Return : public ElementVisitor {
public:
    // Constructor for 'return' object
    Return(){}
    // Applicable if element name is 'return'
    virtual bool applicable(xmlNodePtr node) const{
        return safe_elementName(node) == "return";
    }
    void visit(xmlNodePtr node) const{
        // Output return type
        cout << safe_getAttribute(node, "type");
    }
};

// Class to visit a method
class Method : public ElementVisitor {
private:
    // Indentation level
    int m_tabLevel;
public:
    Method(int tabLevel):
        m_tabLevel(tabLevel){}
    // Applicable if element name is 'method'
    virtual bool applicable(xmlNodePtr node) const{
        return safe_elementName(node) == "method";
    }
    // Method to visit a 'method' element
    virtual void visit(xmlNodePtr node) const{
        // Output indentation
        cout << indent(m_tabLevel);
        // Visit return type
        Return retType;
        accept(node, retType);
        cout << " ";
        // Output method name
        cout << safe_getAttribute(node, "name");
        // Output parameters
        cout << "(";
        Parameter params;
        accept(node, params);
        cout << ")";
        // Output body opener
        cout << " {"
             << endl;
        // Output body
        cout << indent(m_tabLevel + 1);
        Body body;
        accept(node, body);
        // Output body closer
        cout << endl 
             << indent(m_tabLevel) << "}" << endl;
    }
};

// Class to visit a class attribute
class Attribute : public ElementVisitor {
private:
    // Indentation level
    int m_tabLevel;
public:
    // Constructor for 'attribute' object
    Attribute(int tabLevel):
        m_tabLevel(tabLevel){}
    // Applicable if element name is 'attribure'
    virtual bool applicable(xmlNodePtr node) const{
        return safe_elementName(node) == "attribute";
    }
    // Method to visit an 'attribute' element
    virtual void visit(xmlNodePtr node) const{
        // Output indentation
        cout << indent(m_tabLevel);
        // Output attribute declaration
        cout << safe_getAttribute(node, "type")
             << " "
             << safe_getAttribute(node, "name")
             << ";"
             << endl;
    }
};

// Class to visit a 'class' element
class Class : public ElementVisitor{
public:
    // Applicable if element name is 'class'
    virtual bool applicable(xmlNodePtr node) const{
        return safe_elementName(node) == "class";
    }
    // Visit method (Say 'class { ... };')
    virtual void visit(xmlNodePtr node) const{
        // Output class opener
        cout << "class "
             << getName(node) << " {" << endl;
        // Output attributes
        {
            cout << " private:" << endl;
            Attribute attrVisitor(1);
            accept(node, attrVisitor);
        }
        // Output methods
        {
            cout << " public:" << endl;
            Method methVisitor(1);
            accept(node, methVisitor);
        }
        // Output class closer
        cout << "};" << endl;
    }
private:
    // Method to get 'class'-element attribute 'name'
    static string getName(xmlNodePtr node) {
        // Recover 'name' attribute
        return safe_getAttribute(node, "name");
    }
};

// Main routine
int main(int argc, char* argv[]){

    // Runtime check for correct LibXML2 version
    LIBXML_TEST_VERSION

    if(argc != 2){
        cerr << "Usage: example <FILEPATH>" << endl;
        return 1;
    }

    // Parse input file
    xmlDocPtr domRoot = xmlReadFile(argv[1], NULL, 0);

    // Check if parsing was successful
    if(not domRoot){
        cerr << "FAILED: Bad input file" << endl;
        return 1;
    }

    // Recover root node
    xmlNodePtr rootNode = xmlDocGetRootElement(domRoot);

    // Check root node is a 'model'
    if(safe_elementName(rootNode) != "model"){
        cerr << "Root element is not a 'model'" << endl;
        return 1;
    }

    // Attempt to perform passes on DOM
    try{
        Class classVisitor;
        accept(rootNode, classVisitor);
    }
    // Catch exception string
    catch(const string &exception){
        cerr << "FAILED: " << exception << endl;
        return 1;
    }

    // Free document
    xmlFreeDoc(domRoot);

    return 0;
}
